// LDraw quad planarity checker
// Usage: PlanarCheck PlanarCheck [-w <angle>] [-t <angle>] [-c] [-o] LdrawFile [LdrawFileOut]
// Philippe "Philo" Hurbain - 2007 - www.philohome.com
// Contains code from Tomas Mller
// See thesis here: http://www.cs.lth.se/home/Tomas_Akenine_Moller/pubs/tritri.pdf
// Complete code here: http://jgt.akpeters.com/papers/Moller97/tritri.html


#include "stdafx.h"


#define USE_EPSILON_TEST TRUE  
#define EPSILON 0.000001


/* some macros */
#define CROSS(dest,v1,v2)                      \
              dest[0]=v1[1]*v2[2]-v1[2]*v2[1]; \
              dest[1]=v1[2]*v2[0]-v1[0]*v2[2]; \
              dest[2]=v1[0]*v2[1]-v1[1]*v2[0];

#define DOT(v1,v2) (v1[0]*v2[0]+v1[1]*v2[1]+v1[2]*v2[2])

#define SUB(dest,v1,v2) dest[0]=v1[0]-v2[0]; dest[1]=v1[1]-v2[1]; dest[2]=v1[2]-v2[2]; 

#define ADD(dest,v1,v2) dest[0]=v1[0]+v2[0]; dest[1]=v1[1]+v2[1]; dest[2]=v1[2]+v2[2]; 

#define MULT(dest,v,factor) dest[0]=factor*v[0]; dest[1]=factor*v[1]; dest[2]=factor*v[2];

#define SET(dest,src) dest[0]=src[0]; dest[1]=src[1]; dest[2]=src[2]; 

// #define NORM(v1) (sqrt(v1[0]*v1[0] + v1[1]*v1[1] + v1[2]*v1[2]))
#define NORM(v1) (sqrt(DOT(v1,v1)))

void usage(void)
{
	printf ("Usage: PlanarCheck [-w <angle>] [-t <angle>] [-c] [-o] LdrawFile [LdrawFileOut]\n");
	printf ("Options\n");
	printf ("  -w <angle>: Define warning angle threshold\n");
	printf ("  -t <angle>: Define error angle threshold\n");
	printf ("  -c        : Automatically split warped quads into triangles\n");
	printf ("  -o        : Do not create conditional lines when splitting quads\n");
	printf("Press <Enter> to quit");
	getchar();
}

#define ANGLEWARN 1
#define ANGLETHR 3

int main(int argc, char* argv[])
{

	double AngleThr, AngleWarn, DeciAngleWarn, AngleVal1, AngleVal2, AngleMax;

	double InQuad[4][3];

	double a[3], b[3], c[3], d[4];
	double temp[3], n1[3], n2[3];
	int Numline=0;
	int NumBadAngle=0, NumWarnAngle=0, NumDeciWarnAngle=0;

	int i;

	int AutoCorrect=false;
	int NoCond=false;
	int FileOut=false;

	FILE *LdrawFile;
	FILE *LdrawFileOut;


	char buf[1024];
	int type, color;

	printf("\nLdraw quad planarity check v1.4 - by Philo");
	printf("\n------------------------------------------\n\n");

	if(argc < 2) 
	{
		usage();
		return(1);
	}

	AngleWarn=ANGLEWARN;	// Default values are 1 for warning,
	AngleThr=ANGLETHR;		// 3 for error

	for(i=1; i<argc; i++)
	{
		if(argv[i][0] != '-') break;

		switch (argv[i][1])
		{
		case 'c':
			AutoCorrect = true;
			break;
		case 'o':
			NoCond = true;
			break;
		case 'w':
			i++;
			sscanf(argv[i], "%lf", &AngleWarn);
			if (AngleWarn<=0 || AngleWarn >= 90)
			{
				printf("Coplanarity warning angle (-w %.7lg) must be in [0..90]\n", AngleWarn);
				printf("Press <Enter> to quit");
				getchar();
				return 3;
			}
			break;
		case 't':
			i++;
			sscanf(argv[i], "%lf", &AngleThr);
			if (AngleWarn<=0 || AngleWarn >= 90)
			{
				printf("Coplanarity error threshold (-t %.7lg) must be in [0..90]\n", AngleThr);
				printf("Press <Enter> to quit");
				getchar();
				return 3;
			}
			break;
		default:
			usage();
			return 1;
		}
	}
	if((AngleThr != ANGLETHR) || (AngleWarn != ANGLEWARN))
	{
		printf("\n*********************** Warning *************************** \n");
		printf("   Coplanarity angle thresholds use non-standard values\n");
		printf("Results can't be used for LDraw Parts Tracker certification\n");
		printf("*********************** Warning *************************** \n\n");
	}

	if(AngleWarn > AngleThr) 
	{
		AngleWarn=AngleThr;
		printf("Specified warning value > error value\n");
		printf("Using warning angle = %.7lg\n\n", AngleWarn);
	}
	printf("Coplanarity warning angle = %.7lg\n", AngleWarn);
	printf("Coplanarity threshold angle = %.7lg\n", AngleThr);
	DeciAngleWarn=AngleWarn/10;

	
	if(i==argc) 
	{
		usage();
		return(1);
	}

	LdrawFile = fopen (argv[i], "rt");
	if(LdrawFile == NULL)
	{
		printf("Can't open file\n%s\n", argv[i]);
		printf("Press <Enter> to quit");
		getchar();
		return 2;
	}
	i++;
	if(i<argc)
	{
		FileOut=true;
		LdrawFileOut = fopen (argv[i], "wt");
		if(LdrawFileOut == NULL)
		{
			printf("Can't create file\n%s\n", argv[i]);
			printf("Press <Enter> to quit");
			getchar();
			return 2;
		}
	}


	printf("\nProcessing...\n");

	while (fgets(buf, 1024, LdrawFile))
	{
//		if(argc==3) fputs(buf, LdrawFileOut);
		Numline++;
		type = -1;
		sscanf(buf, "%d", &type);
		if(type==4)
		{
			sscanf (buf, "%d %d %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf", &type, &color, 
				&InQuad[0][0], &InQuad[0][1], &InQuad[0][2], 
				&InQuad[1][0], &InQuad[1][1], &InQuad[1][2],
				&InQuad[2][0], &InQuad[2][1], &InQuad[2][2],
				&InQuad[3][0], &InQuad[3][1], &InQuad[3][2]);

			SUB(a,InQuad[0], InQuad[1]);
			SUB(b,InQuad[1], InQuad[2]);
			SUB(c,InQuad[2], InQuad[3]);
			SUB(d,InQuad[3], InQuad[0]);
			CROSS(temp, a ,d);
			MULT(n1,temp,1/NORM(temp));
			CROSS(temp, b ,c);
			MULT(n2,temp,1/NORM(temp));
			CROSS(temp, n1, n2);
			AngleVal1=180/3.1415926*asin((NORM(temp)));

			CROSS(temp, a ,b);
			MULT(n1,temp,1/NORM(temp));
			CROSS(temp, d ,c);
			MULT(n2,temp,1/NORM(temp));
			CROSS(temp, n1, n2);
			AngleVal2=180/3.1415926*asin((NORM(temp)));
			AngleMax=(AngleVal1> AngleVal2) ? AngleVal1:AngleVal2;
			
			if(FileOut)
			{
				if(AngleMax>AngleWarn)
				{
					if(!AutoCorrect) 
					{
						fputs(buf, LdrawFileOut);
						fprintf(LdrawFileOut,"0 Angle %.4lg> %.4lg\n", AngleMax, AngleWarn);
						fprintf(LdrawFileOut, "0 Suggested split:\n");
					}
					if(AngleVal1> AngleVal2)
					{
						if(!AutoCorrect) fprintf(LdrawFileOut, "0 "); 
						fprintf(LdrawFileOut, "3 %d %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", color,
							InQuad[0][0], InQuad[0][1], InQuad[0][2], InQuad[1][0], InQuad[1][1], InQuad[1][2], InQuad[2][0], InQuad[2][1], InQuad[2][2]);
						if(!AutoCorrect) fprintf(LdrawFileOut, "0 "); 
						fprintf(LdrawFileOut, "3 %d %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", color,
							InQuad[0][0], InQuad[0][1], InQuad[0][2], InQuad[2][0], InQuad[2][1], InQuad[2][2], InQuad[3][0], InQuad[3][1], InQuad[3][2]);
						if(!NoCond)
						{
							if(!AutoCorrect) fprintf(LdrawFileOut, "0 ");
							fprintf(LdrawFileOut, "5 24 %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", 
								InQuad[0][0], InQuad[0][1], InQuad[0][2], InQuad[2][0], InQuad[2][1], InQuad[2][2],
								InQuad[1][0], InQuad[1][1], InQuad[1][2], InQuad[3][0], InQuad[3][1], InQuad[3][2]);
						}
					}
					else
					{
						if(!AutoCorrect) fprintf(LdrawFileOut, "0 ");
						fprintf(LdrawFileOut, "3 %d %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", color,
							InQuad[0][0], InQuad[0][1], InQuad[0][2], InQuad[1][0], InQuad[1][1], InQuad[1][2], InQuad[3][0], InQuad[3][1], InQuad[3][2]);
						if(!AutoCorrect) fprintf(LdrawFileOut, "0 ");
						fprintf(LdrawFileOut, "3 %d %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", color,
							InQuad[1][0], InQuad[1][1], InQuad[1][2], InQuad[2][0], InQuad[2][1], InQuad[2][2], InQuad[3][0], InQuad[3][1], InQuad[3][2]);
						if(!NoCond)
						{
							if(!AutoCorrect) fprintf(LdrawFileOut, "0 ");
							fprintf(LdrawFileOut, "5 24 %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", 
								InQuad[1][0], InQuad[1][1], InQuad[1][2], InQuad[3][0], InQuad[3][1], InQuad[3][2],
								InQuad[0][0], InQuad[0][1], InQuad[0][2], InQuad[2][0], InQuad[2][1], InQuad[2][2]);
						}
					}
				}
				else fputs(buf, LdrawFileOut);
			}
			if(AngleMax>AngleThr)
			{
				printf("\nLine %d: Angle = %.4lg", Numline, AngleMax);
				NumBadAngle++;
			}
			else if (AngleMax>AngleWarn) NumWarnAngle++;
			else if (AngleMax>DeciAngleWarn) NumDeciWarnAngle++;
		}
		else if(FileOut) fputs(buf, LdrawFileOut);
	}

	printf("\n\n\n----------File Summary-------------");
	if (NumDeciWarnAngle>0)
		printf("\n\n%d quad(s) between %.4lg and %.4lg degrees", NumDeciWarnAngle, DeciAngleWarn, AngleWarn);

	if (NumWarnAngle>0)
		printf("\n\nWarning: %d quad(s) between %.4lg and %.4lg degrees", NumWarnAngle, AngleWarn, AngleThr);
	else
		printf("\n\nNo quad between %.4lg and %.4lg degrees", AngleWarn, AngleThr);

	if (NumBadAngle>0)
		printf("\n\nERROR: %d quad(s) above %.4lg degrees", NumBadAngle, AngleThr);
	else
		printf("\n\nNo quad above %.4lg degrees", AngleThr);

	printf("\n\n-----------------------------------\n");

	if (NumBadAngle > 0) printf("\n%d warped quad(s) detected",NumBadAngle);
	else if((AngleThr == ANGLETHR) && (AngleWarn == ANGLEWARN))	printf("\nAll quads are planar enough!");
	fclose (LdrawFile);
	if(FileOut)	fclose (LdrawFileOut);

	printf("\n\nPress <Enter> to quit");
	getchar();
	return(0);
}



